
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"
#define _GNU_SOURCE
#include <stdio.h>
#include <alloca.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/types.h>
#include <dirent.h>                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                /*#include <dirent.h> */     /* opendir() closedir() */
#include <fnmatch.h>

#include "../../src/initng_string_tools.h"
#include "../../src/initng_service_cache.h"
#include "../../src/initng_global.h"
#include "../../src/initng_open_read_close.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_plugin_hook.h"

#include "initng_i_parser.h"

static service_h *parse_file(char *filename, char *watch_for);
static service_h *parse_service_line(char **to_parse, char *watch_for,
                                     service_h * father, e_type type,
                                     char *filename);
static int parse_opt(char **where, service_h * srv);

static int set_parser(s_entry * type, char **value, service_h * from_service);
static int unset_parser(s_entry * type, char **value,
                        service_h * from_service);
static int string_parser(s_entry * type, char **value,
                         service_h * from_service);
static int strings_parser(s_entry * type, char **value,
                          service_h * from_service);
static int script_parser(s_entry * type, char **value,
                         service_h * from_service);
static int int_parser(s_entry * type, char **value, service_h * from_service);





/* Load a service from a process_name or process_path */
service_h *initng_i_parser(char *process_name)
{
    char filetoparse[200];
    char *tmp = NULL;
    service_h *got_serv = NULL;

    assert(process_name);
    D_("initng_i_parser(%s);\n", process_name);
    /* example  INITNG_ROOT=/etc/initng process_name="deamon/sshd INITNG_EXT=".i" */
    /* TRY NO 1, try load   "/etc/initng/daemon/sshd.i" */
    /* TRY NO 2, try load   "/etc/initng/daemon/sshd/sshd.i" */
    /* TRY NO 3, try load   "/etc/initng/daemon/sshd/default.i" */
    /* TRY NO 4, try load   "/etc/initng/daemon/default.i" */
    /* TRY NO 5, try load   "/etc/initng/daemon/daemon.i */
    /* TRY NO 6, try load   "/etc/initng/sshd.i" */
    /* TRY NO 7, try load   "/etc/initng/daemon.i */

    /* TRY NO 1, try load   "/etc/initng/daemon/sshd/sshd.i" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, process_name);
    strcat(filetoparse, "/");
    strcat(filetoparse, st_strip_path(process_name));
    strcat(filetoparse, INITNG_EXT);
    if ((got_serv = parse_file(filetoparse, process_name)))
        return (got_serv);

    /* TRY NO 2, try load   "/etc/initng/daemon/sshd.i" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, process_name);
    strcat(filetoparse, INITNG_EXT);
    if ((got_serv = parse_file(filetoparse, process_name)))
        return (got_serv);

    /* TRY NO 3, try load   "/etc/initng/daemon/sshd/default.i" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, process_name);
    strcat(filetoparse, "/default" INITNG_EXT);
    if ((got_serv = parse_file(filetoparse, process_name)))
        return (got_serv);

    /* TRY NO 4, try "/etc/initng/daemon/default.i" */
    tmp = st_get_path(process_name);
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, tmp);
    strcat(filetoparse, "/default" INITNG_EXT);
    free(tmp);
    if ((got_serv = parse_file(filetoparse, process_name)))
        return (got_serv);

    /* TRY NO 5, try "/etc/initng/daemon/daemon.i" */
    tmp = st_get_path(process_name);
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, tmp);
    strcat(filetoparse, "/");
    strcat(filetoparse, tmp);
    strcat(filetoparse, INITNG_EXT);
    free(tmp);
    if ((got_serv = parse_file(filetoparse, process_name)))
        return (got_serv);

    /* TRY NO 6, try "/etc/initng/sshd.i" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, st_strip_path(process_name));
    strcat(filetoparse, INITNG_EXT);
    if ((got_serv = parse_file(filetoparse, process_name)))
        return (got_serv);

    /* TRY NO 7, try "/etc/initng/daemon.i" */
    tmp = st_get_path(process_name);
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, tmp);
    strcat(filetoparse, INITNG_EXT);
    free(tmp);
    if ((got_serv = parse_file(filetoparse, process_name)))
        return (got_serv);



    D_("initng_load_process(%s): got \"NULL\"\n", process_name);
    return (NULL);
}




/* parse a file for services */
static service_h *parse_file(char *filename, char *watch_for)
{
    char *file, *orig;
    service_h *service_to_return = NULL;
    service_h *tmp = NULL;

    assert(filename);
    assert(watch_for);

    D_("parse_file(%s);\n", filename);

    /* read config file */
    if (!open_read_close(filename, &orig))
    {
        D_("Can't open config file %s.\n", filename);
        return (NULL);
    }
    file = orig;


    /* main parse line by line loop */
    while (file[0] != '\0')
    {
        /* skip leading spaces */
        JUMP_SPACES(file);
        if (file[0] == '\0')
            break;

        /* skip lines starting with '#' */
        if (file[0] == '#')
        {
            JUMP_TO_NEXT_ROW(file);
            continue;
        }

        /* if this row has keyword service */
        if (st_cmp(&file, "service"))
        {
            if ((tmp =
                 parse_service_line(&file, watch_for, NULL, TYPE_SERVICE,
                                    filename)))
                service_to_return = tmp;
            /* if this row has keyword class */
        }
        else if (st_cmp(&file, "class"))
        {
            if ((tmp =
                 parse_service_line(&file, watch_for, NULL, TYPE_CLASS,
                                    filename)))
                service_to_return = tmp;
        }
	else if (st_cmp(&file, "daemon"))
	{
            if ((tmp =
                 parse_service_line(&file, watch_for, NULL, TYPE_DAEMON,
                                    filename)))
                service_to_return = tmp;
	}

        JUMP_TO_NEXT_ROW(file);
    }                                       /* end while */

    free(orig);
#ifdef DEBUG_PRINT
    print_all();
#endif
    /*exit_initng(1); */

    /*
     * Search the service chache again for an exeact match,
     * if this wasnt exactly what we where watching for.
     */
    if(!service_to_return || ( service_to_return &&  strcmp(service_to_return->name, watch_for)!=0))
	if((tmp = service_db_find_by_exact_name(watch_for)))
    	    return (tmp);

    /* Return what we got, and hope for the best */
    if (service_to_return)
        return (service_to_return);
	
    return (NULL);
}




static service_h *parse_service_line(char **to_parse, char *watch_for,
                                     service_h * father, e_type type,
                                     char *filename)
{
    char *name = NULL;
    char *father_name = NULL;
    service_h *new_service = NULL;
    service_h *service_to_return = NULL;
    service_h *duplicate = NULL;

    assert(to_parse);
    assert(*to_parse);
    assert(watch_for);

    D_("parse_service_line(*s*,%s);\n", watch_for);

    /* service test : class { */
    /*        |               */
    /* jump forward on spaces */
    JUMP_SPACES(*to_parse);

    /* service test : class { */
    /*         |              */

    /* fetch name */
    if (!(name = st_dup_next_word(to_parse)))
    {
        F_("Did not got name!\n");
        return (NULL);
    }
    /* service test : class { */
    /*             |          */

    D_("parse_service_line(s,%s): service name: \"%s\"\n", watch_for, name);

    /* jump forward on spaces */
    JUMP_SPACES(*to_parse);

    /* service test : class { */
    /*              |         */
    if ((*to_parse)[0] == ':')
    {
        (*to_parse)++;
        /* service test : class { */
        /*               |        */

        /* jump forward on spaces */
        JUMP_SPACES(*to_parse);
        /* service test : class { */
        /*                |       */

        father_name = st_dup_next_word(to_parse);
        if (father_name)
        {
            D_("Father is: %s, set from \"server : father\"\n", father_name);
        }

        /* service test : class { */
        /*                     |  */
        /* jump forward on spaces */
        JUMP_SPACES(*to_parse);

    }

    /* service test : class { */
    /*                      | */
    if ((*to_parse)[0] != '{')
    {
        /* something is wrong */
        F_("error parsing, expected something else.\n");
        return (NULL);
    }

    /* jump to first char after start tag and begin from there. */
    (*to_parse)++;
    /* service test : class { */
    /*                       | */

    /* create a service entry - if there exits an duplicate, merge with this one */
    if ((duplicate = service_db_find_by_exact_name(name)))
        new_service = service_db_copy(name, duplicate);
    else
        new_service = service_db_new(name, type);

    /* check so that it was allocated! */
    if (!new_service)
    {
        F_("Unable to allocate space for new service!\n");
        /* need to free any allocated strings */
        free(name);
        if (father_name)
            free(father_name);
        return (NULL);
    }

    /* set the father to the service */
    if (father_name)
    {
        new_service->father_name = father_name;
        /* father will be service_db_find_by_name(father_name) when needed */
    }
    else if (father && father->name)
    {
        new_service->father_name = father->name;
        new_service->father = father;
    }

    /* why another check? if new_service not exist it will fail upper */
    /*    if (!new_service)
       {
       F_("parse_service_line(s,s): unable to allocate new service!\n");
       free(name);
       if (father_name)
       free(father_name);
       return (NULL);
       }
     */
    if (duplicate)
    {
        /* Service has same name as old service, remove the old one */
        D_("Removing old service for %s\n", new_service->name);
        /* remove it from the service_db */
        list_del(&duplicate->list);
        /* free all the data, and service */
        service_db_free(duplicate);
    }

    /* set type if not set. */
    if (father)
        new_service->type = type;


    /* carry on until segment stop or eof, this will handle all lines in current section */
    while ((*to_parse)[0] != '\0' && (*to_parse)[0] != '}')
    {
        service_h *tmp = NULL;

        while (TRUE)
        {
            /* skip spaces and empty lines with them */
            JUMP_SPACES(*to_parse);
            if (*(to_parse)[0] == '\0' || *(to_parse)[0] == '}')
                break;
            if (*(to_parse)[0] == '\n')
                (*to_parse)++;
            else
                break;
        }

        /* skip lines with '#' */
        if ((*to_parse)[0] == '#')
        {
            JUMP_TO_NEXT_ROW(*to_parse);
            continue;
        }

        /****  Check for subclasses ****/

        /* if this row has keyword service */
        if (st_cmp(to_parse, "service") == TRUE)
        {
            if ((tmp =
                 parse_service_line(to_parse, watch_for, NULL, TYPE_SERVICE,
                                    filename)))
                service_to_return = tmp;
            JUMP_TO_NEXT_ROW(*to_parse);
            continue;
        }

        /* if this row has keyword daemon */
	/*
        if (st_cmp(to_parse, "daemon") == TRUE)
        {
            if ((tmp =
                 parse_service_line(to_parse, watch_for, NULL, TYPE_DAEMON,
                                    filename)))
                service_to_return = tmp;
            JUMP_TO_NEXT_ROW(*to_parse);
            continue;
        }*/

        /* if this row has keyword class */
        if (st_cmp(to_parse, "class") == TRUE)
        {
            if ((tmp =
                 parse_service_line(to_parse, watch_for, NULL, TYPE_CLASS,
                                    filename)))
                service_to_return = tmp;
            JUMP_TO_NEXT_ROW(*to_parse);
            continue;
        }

        /* parse line - search for keywords in g.option_table */
        parse_opt(to_parse, new_service);

        /* ok, this line is parsed, go to next */
        JUMP_TO_NEXT_ROW(*to_parse);
    }

    service_db_set_string(&FROM_FILE, new_service, i_strdup(filename));
    /* add to service db */
    if (!service_db_add(new_service))
    {
        F_("can't add service to service db!\n");
        if (new_service->name)
            free(new_service->name);

        /* remove all data entrys */
        service_db_remove_all(new_service);

        if (new_service)
            free(new_service);
        return (NULL);
    }

    /* check if this is the service we are looking for. */
    if (fnmatch(new_service->name, watch_for, FNM_CASEFOLD) == 0)
        return (new_service);

    /* if we got a service from a under service */
    if (service_to_return)
        return (service_to_return);
    /* return null if not */
    return (NULL);
}



static int parse_opt(char **where, service_h * srv)
{
    s_entry *current = NULL;
    int w_l = 0;
    int result = FALSE;

    while (TRUE)
    {
        /* jump forward all spaces and skip empty or commented lines after that */
        JUMP_SPACES(*where);
        if ((*where)[0] == '\n')
            (*where)++;
        else
            break;
    }

    /* skip empty services */
    if ((*where)[0] == '}')
        return (TRUE);

    /* count number of chars this option name have 
       (option name can contain blank characters - i.e. spaces and tabs */
    while ((*where)[w_l] && (*where)[w_l] != '\n'
           && (*where)[w_l] != ';' && (*where)[w_l] != '=')
        w_l++;

    /* strip ' ' and '\t' from the end */
    while ((*where)[w_l - 1] == ' ' || (*where)[w_l - 1] == '\t')
        w_l--;

    if (!w_l)
    {
        D_("option name to short!\n");
        return (FALSE);
    }

    /* get option name and strip all blanks from it */
    char *opt_name = i_strndup(*where, w_l);

    st_strip(opt_name);

    while_option_db(current)
    {
	/* temporary pointer storage */
	s_entry *tmp = current;
	
	/* make sure opt_name is set */
	if (!current->opt_name) continue;
	
	/* check string lenght of word, this is faster then strcasecmp */
	if ((int) strlen(opt_name) != current->str_len) continue;
	
	/* finaly, check if opt_name is what we are looking for */
	if (strcasecmp(current->opt_name, opt_name) != 0) continue;
	
        D_("parse_opt(%s): option \"%s\" value at option.\n",
               srv->name, current->opt_name);
	       
        /* jump one char */
	*where = *where + w_l;
        /* jump forward all spaces */
        JUMP_SPACES(*where);
	
	/* check that we are not in end of line */
        if (!(*where)[0]) continue;
	
	/* if this is a alias, browse forward to the correct opt_type */
	while (tmp->opt_type==ALIAS && tmp->alias)
	    tmp=tmp->alias;
	
	/* now switch the opt_type */
        switch (tmp->opt_type)
        {
            case STRING:
                result = string_parser(current, where, srv);
                break;
            case STRINGS:
                result = strings_parser(current, where, srv);
                break;
            case SCRIPT:
                result = script_parser(current, where, srv);
                break;
            case SET:
                result = set_parser(current, where, srv);
                break;
            case UNSET:
                result = unset_parser(current, where, srv);
                break;
            case INT:
                result = int_parser(current, where, srv);
                break;
            default:
                F_("Unknowun, or undefined type to parse!\n");
                continue;
        }
    }

    free(opt_name);

    /* return nicely */
    if (result) return(result);


    /* Complain a bit */
    {
        char *line = i_strndup(*where, strcspn(*where, "\n\0"));

        F_("parse_opt(%s): line: \"%s\"!\n", srv->name, line);
        free(line);
        F_("sleeping 2 seconds......\n");
        sleep(2);
    }    
    return(FALSE);
}



static int set_parser(s_entry * type, char **value
                      __attribute__ ((unused)), service_h * from_service)
{
    assert(from_service);
    assert(*value);
    assert(value);

    D_("set_parser(%s,s,%s);\n", type->opt_name, from_service->name);
    service_db_set(type, from_service);
    return (TRUE);
}

static int unset_parser(s_entry * type, char **value __attribute__ ((unused)),
                        service_h * from_service)
{
    assert(from_service);
    assert(*value);
    assert(value);

    D_("unset_parser(%s,s,%s);\n", type->opt_name, from_service->name);
    service_db_remove(type, from_service);
    return (TRUE);
}

static int string_parser(s_entry * type, char **value,
                         service_h * from_service)
{
    char *to;

    assert(from_service);
    if (!(*value))
        return (FALSE);
    D_("string_parser(%s,s,%s);\n", type->opt_name, from_service->name);
    if ((to = st_dup_line(value)))
    {
        service_db_set_string(type, from_service, to);
        return (TRUE);
    }
    return (FALSE);
}

static int int_parser(s_entry * type, char **value, service_h * from_service)
{
    char *to;

    assert(from_service);
    if (!(*value))
        return (FALSE);
    D_("int_parser(%s,s,%s);\n", type->opt_name, from_service->name);
    if ((to = st_dup_line(value)))
    {
        service_db_set_int(type, from_service, atoi(to));
        if (to)
            free(to);
        return (TRUE);
    }
    return (FALSE);
}

static int strings_parser(s_entry * type, char **value,
                          service_h * from_service)
{
    char *to = NULL;

    assert(from_service);
    if (!(*value))
        return (FALSE);
    D_("strings_parser(%s,s,%s);\n", type->opt_name, from_service->name);

    while ((to = st_dup_next_word(value)))
    {
        service_db_set_another_string(type, from_service, to);
        JUMP_SPACES(*value);
    }
    return (TRUE);
    /*return(FALSE); */
}

static int script_parser(s_entry * type, char **value,
                         service_h * from_service)
{
    int stack_length = 0;

    assert(from_service);
    D_("script_parser(%s,s,%s);\n", type->opt_name, from_service->name);
    stack_length = chars_to_end_of_this_stack(*value);

    service_db_set_string(type, from_service,
                          i_strndup(*value, stack_length));
    (*value) = (*value) + stack_length;
    return (TRUE);
}






int module_init(const char *version)
{

    D_("i_parser: module_init();\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module, is compiled for \"%s\" version, and initng is compiled on \"%s\" version, wont load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    return (initng_add_hook(PARSERS, 50, &initng_i_parser));

}

void module_unload(void)
{
    D_("i_parser: module_unload();\n");
    initng_del_hook(PARSERS, &initng_i_parser);
}
