
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/poll.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <assert.h>

#include "../../src/initng_global.h"
#include "../../src/initng_process_db.h"
#include "../../src/initng_service_cache.h"
#include "../../src/initng_handler.h"
#include "../../src/main.h"
#include "../../src/initng_active_db.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_load_module.h"
#include "../../src/initng_plugin_callers.h"
#include "../../src/initng_global.h"
#include "../../src/initng_error.h"
#include "../../src/initng_string_tools.h"

#include "initng_unneeded.h"


static int cmd_stop_unneeded(char *arg);
s_command STOP_UNNEEDED =
    { 'y', "stop_unneeded", TRUE_OR_FALSE_COMMAND, NO_OPT,
    {(void *) &cmd_stop_unneeded},
    "Stop all services, not in runlevel"
};

static int cmd_stop_unneeded(char *arg)
{

    active_h *needed[200];
    int i,j, k;
    active_h *s, *q = NULL;

    S_;
    
    /* clear the array */
    for(i=0;i<200;i++)
	needed[i]=NULL;

    /* Recursively mark all dependencies of the current runlevel */
    if (!(s = active_db_find_by_name(g.runlevel)))
        return(FALSE);
    
    /* mark the service a a needed one, we walk out from this. */
    needed[0]=s;

    i=0;
    while(i<200 && needed[i])
    {
	D_("service: %s depends on: ", needed[i]->name);
	active_h **deps=active_db_dep_on(needed[i]);
	j=0;
	/* fill the need with these deps */
	while(deps[j])
	{
	    k=0;
	    /* write over if it exists allredy */
	    while(needed[k] && needed[k]!=deps[j] && k<200)
		k++;
	    needed[k]=deps[j];
	    
	    printf(" %s", deps[j]->name);
	    j++;
	}
	free(deps);
	i++;
	printf("\n");
    }
    
#ifdef DEBUG
    for(i=0;i<200 && needed[i]; i++)
     D_("Wont stop %s\n", needed[i]->name);
#endif

    /* Now rewalk the tree, stopping anything not marked as needed */
    s = NULL;
    while_active_db_safe(s, q)
    {
	/* if this service exits in the need db, continue */
	i=0;
	while(i<200 && needed[i]) {
	    if(needed[i]==s)
        	i=10000;
	    i++;
	}
	if (i>200) continue;
        
	/* dont stop virutals */
	if (s->type == TYPE_VIRTUAL) continue;
	
	W_("Will stopp %s\n", s->name);
	stop_service(s);
    }
    
    return(TRUE);
}




int module_init(const char *version)
{
    D_("module_init(unneeded);\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module, is compiled for \"%s\" version, and initng is compiled on \"%s\" version, wont load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    initng_command_add(&STOP_UNNEEDED);

    D_("libunneeded.so.0.0 loaded!\n");
    return (TRUE);
}


void module_unload(void)
{
    D_("module_unload(unneeded);\n");

    initng_command_del(&STOP_UNNEEDED);


    D_("libunneeded.so.0.0 unloaded!\n");

}
