/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <dlfcn.h>
#include <assert.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>                          /* fcntl() */
#include <time.h>
#include "initng.h"
#include "initng_service_cache.h"
#include "initng_load_module.h"
#include "initng_toolbox.h"
#include "initng_global.h"
#include "initng_signal.h"

#include "initng_plugin_callers.h"
#include "initng_plugin_type.h"

/* P L U G I N    C A L L E R S */

#ifdef DEBUG
/* used to get verbose debug msgs right */
static const char *get_type_string(e_ptype type)
{
    switch (type)
    {
        case EMPTY_UNKNOWN_TYPE:
            return ("NOPLUG");
        case ASTATUS_CHANGE:
            return ("ASTATUS_CHANGE");
	case CONFIRM_ASTATUS_CHANGE:
	    return ("CONFIRM_ASTATUS_CHANGE");
        case PARSERS:
            return ("PARSERS");
        case SWATCHERS:
            return ("SWATCHERS");
        case FDWATCHERS:
            return ("FDWATCHERS");
        case PIPEWATCHERS:
            return ("PIPEWATCHERS");
        case TIMERS:
            return ("TIMERS");
        case STARTL:
            return ("STARTL");
        case STOPL:
            return ("STOPL");
        case MAIN:
            return ("MAIN");
        case A_FORK:
            return ("A_FORK");
        case HANDLE_KILLED:
            return ("HANDLE_KILLED");
        case COMPENSATE_TIME:
            return ("COMPENSATE_TIME");
	case ERR_MSG:
	    return ("ERR_MSG");
    }
    return (NULL);
}
#endif


/* this function returns first pointer got */
void *initng_plugin_pcall(e_ptype type, void *data_args)
{
    void *data_ret = NULL;

    s_call *current, *q = NULL;

    while_call_db_safe(current, q)
    {
        if (current->ptype == type)
        {
            D_("Calling ptype %s plugin from %s\n", get_type_string(type),
               current->from_file);
            if ((data_ret = (*current->c.phook) (data_args)))
            {
                return (data_ret);
            }
        }
    }
    return (NULL);
}

/* not editing anything, so it doesn't have to use a recursive function */
void initng_plugin_timers(void)
{
    time_t current_time = time(NULL);
    s_call *current, *q = NULL;

    while_call_db_safe(current, q)
    {
        if (current->ptype == TIMERS
            && (current->c.tdh->next_hit <= current_time))
        {
            D_("Calling timer (%d, %d)\n", current->c.tdh->next_hit,
               current->c.tdh->seconds);
            current->c.tdh->next_hit = time(NULL) + current->c.tdh->seconds;
            alarm(current->c.tdh->seconds);
            D_("Calling timer plugin from %s\n", current->from_file);
            (*current->c.tdh->call_module) ();
        }
    }
}

int initng_plugin_handle_killed(active_h * s, process_h * p)
{
    s_call *current, *q = NULL;

    while_call_db_safe(current, q)
    {
        if (current->ptype != HANDLE_KILLED) continue;
        
        D_("Calling killed_handle plugin from %s\n", current->from_file);
        if (((*current->c.handle_killed) (s, p)) == TRUE)
            return (TRUE);
        
    }
    return (FALSE);
}


void initng_compensate_time_call(int t)
{
    s_call *current, *q = NULL;

    while_call_db_safe(current, q)
    {
        if (current->ptype == COMPENSATE_TIME && current->c.swatcher)
        {
            D_("Calling system_state_changed plugin from %s\n",
               current->from_file);
            (*current->c.swatcher) (t);
        }
    }

}

/* called when system state has changed. */
void initng_load_module_system_changed(h_sys_state state)
{
    s_call *current, *q = NULL;

    while_call_db_safe(current, q)
    {
        if (current->ptype == SWATCHERS && current->c.swatcher)
        {
            D_("Calling system_state_changed plugin from %s\n",
               current->from_file);
            (*current->c.swatcher) (state);
        }
    }
    return;
}
