
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#define _GNU_SOURCE
#include <fnmatch.h>

#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include "initng_string_tools.h"
#include "initng_toolbox.h"

/*
 * this splits an long string, with separators, like \n , ; ' '
 * into an char **
 * buf: the line to split.
 * len; An pointer sent back with lenght
 * plus: Allocate this mutch extra char ** size;
 * ofs: (offset) start a bit into the buf
 *
 * OBS, this functions only creats a pointer list, that links
 * to buf, if buf is freed, the list dies!
 */
char **split_n(char *buf, int *len, int plus, int ofs)
{
    int n = 1;
    char **v = 0;
    char **w;

    /* step 1: count tokens */
    char *s;

    for (s = buf; *s; s++)
        if (*s == '\n' || *s == ',' || *s == ';' || *s == ' ')
            n++;
    /* step 2: allocate space for pointers */
    v = (char **) initng_calloc((n + plus), sizeof(char *));
    if (!v)
        return 0;
    w = v + ofs;
    *w++ = buf;
    for (s = buf;; s++)
    {
        while (*s && *s != '\n' && *s != ',' && *s != ';' && *s != ' ')
            s++;
        if (*s == 0)
            break;
        if (*s == '\n' || *s == ',' || *s == ';' || *s == ' ')
        {
            *s = 0;
            *w++ = s + 1;
        }
    }
    *len = w - v;
    return v;
}

/*
 * a stack starts with {, and this function calculates
 * the string lent to the }, there can be many
 * { and } in the middle.
 */
int chars_to_end_of_this_stack(char *string)
{
    int i = 0;
    int stacks = 1;

    while (string[i] != '\0')
    {
        if (string[i] == '{')
            stacks++;
        if (string[i] == '}')
        {
            stacks--;
            if (stacks == 0)
                return (i);
        }
        i++;
    }
    return (FAIL);
}

/*
 * a simple strncasecmp, thats only returns TRUE or FALSE
 * it will also add the pointer of string, the value, to forward the
 * pinter to next word.
 */
int st_cmp(char **string, const char *to_cmp)
{
    int chars = strlen(to_cmp);

    /* skip beginning first spaces */
    JUMP_SPACES(*string)
        /* this might be an "comp pare" */
        if ((*string)[0] == '"' && to_cmp[0] != '"')
        (*string)++;

    /* ok, strcasecmp this */
    if (strncasecmp((*string), to_cmp, chars) == 0)
    {
        /* increase the string pointer */
        (*string) += chars;
        return (TRUE);
    }
    return (FALSE);
}

/*
 * Senses what in the string, that is a word, (can be a
 * line with "this is one word", it will i_strdup it, and
 * add word lenght on string pointer
 */
char *st_dup_next_word(char **string)
{
    char *td = NULL;
    int i = 0;

    /* skip beginning first spaces */
    JUMP_SPACES(*string);

    /* count tokens */
    if (*(string)[0] == '"')
    {
        (*string)++;
        while ((*string)[i] != '\0' &&
               (*string)[i] != '"' &&
               (*string)[i] != '\n' &&
               (*string)[i] != ',' && (*string)[i] != ';')
            i++;
    }
    else
        while ((*string)[i] != '\0' &&
               (*string)[i] != ' ' &&
               (*string)[i] != '\t' &&
               (*string)[i] != '\n' && (*string)[i] != ','
               && (*string)[i] != ';')
            i++;

    if (!i)
        return (NULL);

    /* copy string */
    td = i_strndup(*string, i);
    (*string) += i;

    if ((*string)[0] == '"')
        (*string)++;
    return (td);
}

/*
 * This will dup, in the string, what is resolved as the line,
 * (stops with '\n', ';', '\0'), copy it and increment *string with len
 */
char *st_dup_line(char **string)
{
    char *td = NULL;
    int i = 0;

    /* count tokens */
    if ((*string)[0] == '"')
    {
        (*string)++;
        while ((*string)[i] != '\0' &&
               (*string)[i] != '"' &&
               (*string)[i] != '\n' && (*string)[i] != ';')
            i++;
    }
    else
        while ((*string)[i] != '\0' &&
               (*string)[i] != '\n' && (*string)[i] != ';')
            i++;

    if (!i)
        return (NULL);

    /* copy string */
    td = i_strndup(*string, i);
    (*string) += i;

    /* if this is a "line", dont leave an '"' */
    if ((*string)[0] == '"')
        (*string) += i;

    return (td);
}


/* function strips "test/go/th" to "th" */
char *st_strip_path(char *string)
{
    int i = 0;

    /* go to end */
    while (string[i] != '\0')
        i++;

    /* go back to the '/' char */
    while (i && string[i] != '/')
        i--;

    if (string[i] == '/')
        return (&string[i + 1]);
    return (string);
}

/* function strips "test/go/th" to "test/go" */
char *st_get_path(char *string)
{
    int i = 0;

    /* go to end */
    while (string[i] != '\0')
        i++;
    /* step back */
    while (i && string[i] != '/')
        i--;

    if (i)
        return (i_strndup(string, i));

    return (i_strdup(string));
}

/* function strips "test/go/th" to "test/go" and then to "test" */
int st_strip_end(char **string)
{
    int i = 0;

    /* stand on last char */
    while ((*string)[i] != '\0')
        i++;

    /* rewind util standing on a '/' */
    while (i && (*string)[i] != '/')
        i--;

    /* if we really are standing on a '/' */
    if ((*string)[i] == '/')
    {
        (*string)[i] = '\0';                /* truncate */
        return (TRUE);
    }
    return (FALSE);
}

/* function strips all blanks from string */
int st_strip(char *string)
{
    int i = 0, j = 0;

    if (!string)
        return (FALSE);

    while (string[i] != '\0')
    {
        if (string[j] != ' ' && string[j] != '\t')
            j++;
        i++;
        string[j] = string[i];
    }

    return (TRUE);
}
